use clap::{arg, ArgMatches, Command};
use icepick_coin::{Coin, Unsupported};
use solana_sdk::{instruction::Instruction, pubkey::Pubkey, system_instruction};
use std::str::FromStr;

pub struct SOL;

impl Coin for SOL {
    fn coin_name(&self) -> &'static str {
        "sol"
    }

    fn coin_description(&self) -> &'static str {
        "Bring blockchain to the people. Solana supports experiences for power users, new consumers, and everyone in between."
    }

    fn build_transfer_command(&self, command: Command) -> Result<Command, Unsupported> {
        Ok(command
           .arg(arg!(--"fee-payer" [FEE_PAYER] "Use an alternative derivation account for paying fees."))
           .arg(arg!(--"blockhash" <BLOCKHASH> "A recent blockhash to include in the Transaction")))
    }

    fn build_stake_command(&self, _command: Command) -> Result<Command, Unsupported> {
        Err(Unsupported)
    }

    fn run_transfer_command(&self, matches: &ArgMatches) -> Result<(), Box<dyn std::error::Error>> {
        let amount = matches
            .get_one::<String>("amount")
            .and_then(|amount| u64::from_str(amount).ok())
            .expect("amount: valid int");
        let to_address = matches
            .get_one::<String>("to-address")
            .and_then(|pk| Pubkey::from_str(pk).ok())
            .expect("to-address: valid public key");
        let from_account = matches
            .get_one::<String>("from-account")
            .and_then(|index| u32::from_str(index).ok())
            .filter(|index| *index < (1 << 31));
        let fee_payer = matches.get_one::<String>("fee-payer");
        let blockhash = matches.get_one::<String>("blockhash");
        // We need: pubkey 1, pubkey 2, lamports,
        let instruction = system_instruction::transfer(&to_address, &to_address, amount);
        print_instruction(&instruction);
        Ok(())
    }
}

fn guess_instruction(i: &Instruction) -> Option<String> {
    let data = &i.data;
    let instruction = data[..4]
        .try_into()
        .ok()
        .map(u32::from_le_bytes)?;
    let system_program = solana_sdk::system_program::ID;

    let entries = [(system_program, 2u32, |data: &[u8]| {
        let lamports = data[..8]
            .try_into()
            .ok()
            .map(u64::from_le_bytes)
            .expect("unable to parse data into lamport amount");
        format!("transfer {lamports} lamports")
    })];

    entries
        .iter()
        .find(|entry| entry.0 == i.program_id && entry.1 == instruction)
        .map(|entry| entry.2(&data[4..]))
}

fn print_instruction(instruction: &Instruction) {
    println!("Program ID: {}", instruction.program_id);
    if !instruction.accounts.is_empty() {
        println!("Associated Accounts:");
        for account in &instruction.accounts {
            println!("|- Public Key: {}", account.pubkey);
            println!("|- Is Signer: {}", account.is_signer);
            println!("\\- Is Writable: {}", account.is_writable);
        }
        match guess_instruction(instruction) {
            Some(guess) => {
                println!("Instruction: {guess}");
            }
            None => {
                println!("Unknown instruction: {:?}", instruction.data);
            }
        }
    }
}
