use clap::command;
use icepick_module::help::*;
use std::{
    io::{IsTerminal, Write},
    process::{Command, Stdio},
};

pub fn guessed_bins() -> Vec<String> {
    "sol".split_whitespace().map(|s| s.to_owned()).collect()
}

pub fn get_command(bin_name: &str) -> (&str, Vec<&str>) {
    if std::env::vars().any(|(k, _)| &k == "ICEPICK_USE_CARGO") {
        ("cargo", vec!["run", "--bin", bin_name, "--"])
    } else {
        (bin_name, vec![])
    }
}

pub fn do_cli_thing() {
    let mut commands = vec![];
    let mut icepick_command = command!();
    for coin_bin in guessed_bins() {
        // try to run the "help" operation on any bins.
        let bin = format!("icepick-{coin_bin}");
        let (command, args) = get_command(&bin);
        let mut child = Command::new(command)
            .args(args)
            .stdin(Stdio::piped())
            .stdout(Stdio::piped())
            .spawn()
            .unwrap();
        let mut input = child.stdin.take().unwrap();
        input
            .write_all("{\"operation\": \"help\"}\n".as_bytes())
            .unwrap();
        input
            .write_all("{\"operation\": \"exit\"}\n".as_bytes())
            .unwrap();
        let output = child.wait_with_output().unwrap().stdout;
        let operations: Vec<Operation> = serde_json::from_slice::<Vec<Operation>>(&output)
            .expect("successful deserialization of operation");
        commands.push((coin_bin, operations));
    }
    let commands = commands.leak();
    for command in commands.iter() {
        let mut subcommand = clap::Command::new(command.0.as_str());
        for op in &command.1 {
            let mut op_command = clap::Command::new(&op.name).about(&op.description);
            for arg in &op.arguments {
                let mut op_arg = clap::Arg::new(&arg.name).help(arg.description.as_str());
                op_arg = match arg.r#type {
                    ArgumentType::Required => op_arg.required(true),
                    ArgumentType::Optional => op_arg.required(false).long(&arg.name),
                };
                op_command = op_command.arg(op_arg);
            }
            subcommand = subcommand.subcommand(op_command);
        }
        icepick_command = icepick_command.subcommand(subcommand);
    }

    let stdin = std::io::stdin();
    let mut cli_input: Option<serde_json::Value> = None;
    // HACK: Allow good UX when running from CLI without piping in a blob.
    // This is because we have CLI arguments _only_ for the module. We _could_ add a global
    // argument `--input` instead of reading input.
    if !stdin.is_terminal() {
        cli_input = serde_json::from_reader(stdin).ok();
    }
    dbg!(&cli_input);
    let blob = cli_input.as_ref().and_then(|json| {
        json.get("blob")
    });

    let matches = icepick_command.get_matches();
    if let Some((module, matches)) = matches.subcommand() {
        if let Some((subcommand, matches)) = matches.subcommand() {
            if let Some(operation) = commands
                .iter()
                .find(|(name, _)| name == module)
                .and_then(|(_, operations)| operations.iter().find(|o| o.name == subcommand))
            {
                let mut args = std::collections::HashMap::<&String, Option<&String>>::with_capacity(
                    operation.arguments.len(),
                );
                for arg in &operation.arguments {
                    args.insert(&arg.name, matches.get_one::<String>(&arg.name));
                }
                let json = serde_json::json!({
                    "operation": subcommand,
                    "values": args,
                    "derived-keys": [],
                    "blob": blob,
                });
                let bin = format!("icepick-{module}");
                let (command, args) = get_command(&bin);
                let mut child = Command::new(command)
                    .args(args)
                    .stdin(Stdio::piped())
                    .stdout(Stdio::piped())
                    .spawn()
                    .unwrap();
                let mut input = child.stdin.take().unwrap();
                serde_json::to_writer(&mut input, &json).unwrap();
                input.write_all(b"\n{\"operation\": \"exit\"}\n").unwrap();
                let output = child.wait_with_output().unwrap().stdout;
                let json: serde_json::Value = serde_json::from_slice(&output).expect("valid json");
                let json_as_str = serde_json::to_string(&json).unwrap();
                println!("{json_as_str}");
            }
        }
    }
}
